/**********************************************************************\
|								       |
|				   				       |
|                        Temporal Access Method:	               |
|                     	  Overlapping B+-trees                         |
|		          (external interface)            	       |
|		                                        	       |
|   	 	      Last modified: October 8, 2000         	       |
|								       |
\**********************************************************************\

			    COPYRIGHT NOTICE

This source code was created by Mr. Theodoros Tzouramanis
<theo@delab.csd.auth.gr>, from Aristotle University of Thessaloniki,
Department of Informatics, Thessaloniki, Greece,
under the supervision of Prof. Yannis Manolopoulos and
Prof. Nikos Lorentzos.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings:
a. Any copy made of this software must include this copyright notice
   in full.
b. All materials developed as a consequence of the use of this
   software shall duly acknowledge such use, in accordance with the usual
   standards of acknowledging credit in academic research.
c. The authors have made no warranty or representation that the
   operation of this software will be error-free or suitable for any
   application, and they are under no obligation to provide any
   services, by way of maintenance, update, or otherwise.  The software
   is an experimental prototype offered on an as-is basis.
d. Redistribution for profit requires the express, written permission
   of the authors. */


#ifndef _overbt_h
#define _overbt_h

/*~~~~~~~~~~~~~~~~~~~~~~~~~ files to include ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#include <stdio.h>

const
 base       = 1;

#define ARRAYsize  5
#define MINin      125
#define MAXin      250            /* internal node capacity of T-tree */
#define MINex      2
#define MAXex      4              /* leaf node capacity of T-tree */
#define WMAXin     2*MAXin+MINin
#define WMAXex     2*MAXex+MINex

#define POINTER   0

#define NULLPOS   -1
#define MAXINT    0x7FFF
#define MAXLONG   0x7FFFFFFFL

#ifndef TRUE
 #define FALSE 0
 #define TRUE  ! FALSE
#endif

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ T-tree ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/*~~~~~~~~~~~~~~~~ special header entry for internal node ~~~~~~~~~~~~~*/

typedef struct inner {      /* internal node of the T-tree */
  long  key[MAXin+1];
  int   p[MAXin+1] ;
} Inner;

/*~~~~~~~~~~~~~~~~~~~~ special header entry for leaf node ~~~~~~~~~~~~~*/

typedef struct leaf {       /* leaf node of the T-tree */
  long  key[MAXex+1];       /* key[0] = maxkeyinnode, seckey[0] is of no use */
  int   seckey[MAXex+1];    /* salary */
  int   endtime;
  int   Bpoint,
	BCpoint,
	Fpoint,
	FCpoint;
} Leaf;

typedef  union  {
	Inner In;
	Leaf  Ex;
} Node;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~  T-tree node  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~ (the form of a node of the T-tree) ~~~~~~~~~~~~~~~~~~~~*/

typedef struct page {
	 int  pos;
	 int  data;         /* boolean, 1:leaf ; 0:internal */
	 int  starttime;
	 int  contain;      /* the number of records in the node */
	 Node node;
       } BTREE;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*\
|	The structures Inner and Leaf are all identical in size.
|       Through the union Node node, it is possible to access the
|       first space in any node	as: node.In and node.Ex, depending
|       on the algorithms' needs.  The value of the status flag
|       isLEAF should always determine how the first node space is
|       used.  The node structure is defined below.
|
|
|					   Leaf node:
|
|	                        	+---------------+
|					|Fpoint |	|
|	                                |       |       |
|					|       |FCpoint|
|	  		       		|---------------|
|	  				|	 	|
|	 Internal node:			|   starttime	|
|		 			|	 	|
|	+---------------+		|---------------|
|	|	 	|		|	 	|
|	|   starttime   |               |    endtime    |
|	|	 	|		|	 	|
|	|---------------|		|---------------|
|	|	 	|		|	 	|
|	|   pos(ition 	|		|  pos(ition 	|
|	|   in file) 	|		|   in file) 	|
|	|---------------|		|---------------|
|	|	 	|		|	 	|
|	|     data 	|		|     data 	|
|	|        	|		|	 	|
|	|---------------|		|---------------|
|	|	 	|		|	 	|
|	|    contain    |		|    contain 	|
|	|        	|		|	 	|
|	+---------------+		+---------------+
|	|    key[0]	|		|    key[0]	|
|	|      		|		|      or	|
|	|     p[0]   	|		| maxkeyinnode()|
|	|---------------|		|---------------|
|	|    key[1]    	|		|    key[1]    	|
|	|     		|		|               |
|	|     p[1]    	|		|   seckey[1]	|
|	|---------------|		|---------------|
|	|	.	|		|	.	|
|	|	.	|		|	.	|
|	|	.	|		|	.	|
|	|---------------|		|---------------|
|	|  key[MAXin]  	|		|   key[MAXex]	|
|	|     		|		|     		|
|	|   p[MAXin]    |       	| seckey[MAXex]	|
|	+---------------+		+---------------+
|					|Bpoint |	|
|	                                |       |       |
|					|       |BCpoint|
|	                        	+---------------+
|
\*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/*---------------------------------------------------------------------*\
|   Implementation Hiding Macro Definitions for B+-Tree nodes		|
\*---------------------------------------------------------------------*/

#define internal(j)         (j).node.In
#define external(j)         (j).node.Ex
#define isLEAF(j)           (j).data
#define maxkeyfor(j)        (j).node.Ex.key[0]    /* key[0] = maxkeyinnode*/

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~ end of T-tree ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct    item {
	 long int key;
	 int      p;
	 int      seckey;
       } item;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ workspace ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~ please consult the reference [LD86] (details in the readme.txt) ~~~~*/

/*~~~~~~~~~ special header entry for internal workspace node ~~~~~~~~~~~~~*/

typedef struct winner {
  long     key[WMAXin+1];
  int      p[WMAXin+1];
  long int savedkey;
} WInner;

/*~~~~~~~~~~~~~~~ special header entry for workspace leaf node ~~~~~~~~~~~*/

typedef struct wleaf {
  long  key[WMAXex+1];
  int   seckey[WMAXex+1];
  int   rentpos;            /* rent position in the file for future use */
  int   savedbpoint;        /* points to its historical leaf node */
} WLeaf;

typedef  union  {
	WInner WIn;
	WLeaf  WEx;
} WorkspcNode;

/*~~~~~~~~~~~~~~~~~~~~~~~~~ workspace node ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct workspace {
	 int         data ;       /* boolean */
	 int         recontain;   /* the real number of keys */
	 int  	     contain;
	 WorkspcNode Wnode;
       } Workspace;

/*---------------------------------------------------------------------*\
|   Implementation Hiding Macro Definitions for workspace nodes		|
\*---------------------------------------------------------------------*/

#define inter(j)  (j).Wnode.WIn
#define exter(j)  (j).Wnode.WEx

/*~~~~~~~~~~~~~~~~~~~~~~~~~ end of workspace ~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct stack {          /* the form of a node of the TTT structure */
	 int pos;
	 int transaction;       /* time point */
	 int treepointer;       /* a T-tree root */
       } STACK;

typedef struct {long key; char tag;} REC;
typedef struct link {REC rec; struct link *next;} SL;
typedef struct stacklink {STACK mainroot; struct stacklink *prev, *next;} STACKLINK; /* The TTT structure is a double linked list*/

int          CountRecpos,         /* the position counter of the T-tree binary file */
	     CountTranspos;       /* the position counter of the TTT binary file */
Workspace    w[ARRAYsize];
long int     WritenNodes, ReadNodes;              /* parameters used only in experimentation */
int          WritenTrans, ReadTrans;              /* parameters used only in experimentation */
int          PrintedKeys, NodesInTree, LeafNodes; /* parameters used only in experimentation */

#define   NOTFIRSTTRANS    CountTranspos!=1

/*---------------------------------------------------------------------*\
|      	                 Database  Utilities				|
\*---------------------------------------------------------------------*/

void  makelist(SL **f, FILE *file, int MAXNumKeys);
//SL    *getmemSL();
void  push(REC a, SL **f);
void  insAfter(SL **node, REC a);
void  inOrder(REC a, SL **f);
REC   getnextkey(SL **f,int *flag);
void  getmemForSTACK(STACKLINK **f, STACKLINK **m);
void  initSTACKlist(STACKLINK **f, STACKLINK **m,FILE *TTTfile);
void  freeSTACKlist(STACKLINK *f);
void  initCounters(int *CountforTrans, int *CountforRecs);

/*----------------------Overlapping B+-tree methods----------------------*/

void  batch_modify(SL **first, int *rpos, FILE *RECfile);
void  BTree_path(Workspace w[], long x, int *height, BTREE buffer[], FILE *RECfile);
void  bin_searchinW(Workspace  w, long newrec, int *pos, int *found);

void  insertBTREEnodeinW(Workspace *w,BTREE q);
void  restoreBTREEnodefromWi(BTREE *p, Workspace *w,int numKeys, item *k);
void  newBTreenode(BTREE *r, Workspace *w, int numKeys, item *k);
void  modifykinWh(Workspace *w, REC k);

long  kfieldinw(BTREE buffer[], int height);
void  setpointers(BTREE *bp, Workspace *w);
void  emptybuffer(BTREE buffer, FILE *RECfile);
int   isDifferent(Workspace w,BTREE b);

void  insertRECinW(Workspace *w, item g);
void  removeRECinW(Workspace *w, item g);
void  updateRECinW(Workspace *w, item g);

void  split(int lev,Workspace w[],BTREE buffer[], FILE *RECfile, int fromfree);
BTREE restore(Workspace *w, int *qpos, long lastkinW, BTREE *buffer, FILE *RECfile);
void  freeW(int *lev, Workspace w[], BTREE buffer[], long lastk,
	    int *rpos, int *LEAFmerge, FILE *RECfile);

/*-------------------------------------------------------------------------*/

void  bin_search(BTREE a, long newrec, int *pos, int *found);
STACK newtreeroot( void );
void  printtree(int ppos, int *lab, FILE *RECfile, FILE *PRINTfile);

void  locate(FILE *RECfile, long rl, long rpos);
void  writeBTreeNode(FILE *RECfile, BTREE t);
BTREE getBTreeNode(FILE *RECfile, long rpos);

void  writeStackNode(FILE *TTTfile, STACK t);
STACK getStackNode(FILE *TTTfile, long rpos);

STACKLINK *findtrans(STACKLINK *l,int no);
void      putRootpos(int rootpos, FILE *TTTfile);
int       getRootpos(FILE *RECfile);

/*-------------------------------------------------------------------------*/

FILE  *openWfile(char file[]);
FILE  *openRWfile(char file[]);
FILE  *openRTfile(char file[]);
FILE  *openWTfile(char file[]);

/*-------------------------------------------------------------------------*/

void  search(STACKLINK *head, STACKLINK *last, FILE *RECfile, FILE *PRINTfile);
int   findLEAFnode(long x, int rootpos, BTREE *temp, int *pos, FILE *RECfile);
void  TimesliceQuery(char ch, STACKLINK *first, FILE *RECfile, FILE *PRINTfile);
void  printSliceTree(int ppos, long x1, long x2, FILE *RECfile);
void  PurekeyQuery(char ch, STACKLINK *first, STACKLINK *last, FILE *RECfile,
		   FILE *PRINTfile);
void  PurekeyQueryWithPredictTime(char ch, STACKLINK *first, STACKLINK *last,
		   FILE *RECfile, FILE *PRINTfile);
void  KindOfPurekeyQuery( char ch, STACKLINK *first, STACKLINK *last,
		   int trans1, int trans2, FILE *RECfile, FILE *PRINTfile);
void  printkey( int ii, BTREE p, long x, int pos,int found,
		   int numberkeys, FILE *PRINTfile);
void  historywithFpoints( STACKLINK *first, STACKLINK *last, long x,int trans1,
		   int trans2, int numberkeys, FILE *RECfile, FILE *PRINTfile);
void  findrestKeysUsingFpoints(BTREE tempnode, long x, int ii, int rootpos,
		   int sumaskedkeys, FILE *RECfile, FILE *PRINTfile);
void  historywithBpoints( STACKLINK *first, STACKLINK *last, long x,int trans1,
		   int trans2, int numberkeys, FILE *RECfile, FILE *PRINTfile);
void  findrestKeysUsingBpoints(BTREE tempnode, long x, int ii,int rootpos,
		   int sumaskedkeys, FILE *RECfile, FILE *PRINTfile);

/*-------------------------------------------------------------------------*/

#endif